/***************************************************************************
  polarized_light_encoder.ino
  Code by lingib https://www.instructables.com/
  Last update 3 February 2023

 ------------
 About
 ------------
 This software decodes the output from my polarized light.
 
 A digital filter is used to smooth the output waveforms from the phototransistors.
 The digital filter code belongs to Curio Res
 https://www.youtube.com/watch?v=eM4VHtettGg&list=PL3BbZZ9thFk5eGKN0-Xo64vwgZQk_sJmB&index=4

 Two sine waves are output for each revolution of the motor shaft
 The code for counting the transitions belongs to Cuirio Res and is explained here
 https://curiousscientist.tech/blog/as5600-magnetic-encoder-a-practical-example

 All other code is mine

  ----------
  COPYRIGHT
  ----------
  This code is free software: you can redistribute it and/or
  modify it under the terms of the GNU General Public License as published
  by the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This software is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License. If
  not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

// ----- Bit set/clear/check/toggle macros
#define SET(x,y) (x |=(1<<y))
#define CLR(x,y) (x &= (~(1<<y)))
#define CHK(x,y) (x & (1<<y))
#define TOG(x,y) (x^=(1<<y))

// ----- constants
#define PI 3.1415926535897932384626433832795
#define HALF_PI 1.5707963267948966192313216916398
#define TWO_PI 6.283185307179586476925286766559
#define DEG_TO_RAD 0.017453292519943295769236907684886
#define RAD_TO_DEG 57.295779513082320876798154814105
#define CW true
#define CCW false

// ----- ADC definitions
const byte ADC0 = A0;
const byte ADC1 = A1;

// ----- motor controller definitions
const byte DIR1 = 9;                                    // PB1 (PORTB)
const byte STEP1 = 8;                                   // PB0 (PORTB)

// ----- motor definitions
#define STEPS_PER_MM  (200*16)/40                       // 200steps/rev; 16 x microstepping; GT2-20 pulley circumference
unsigned int PULSE_WIDTH = 2;                           // easydriver step pulse-width (uS)
unsigned long DELAY_MIN = 250L;                        // minimum inter-step delay (uS) between motor steps (controls speed)
bool DIRECTION1 = CW;                                   // current motor directions

// ----- Auto-calibration
float
Xmax,
Xmin,
Xmid,
Xpeak;

float
Ymax,
Ymin,
Ymid,
Ypeak,
Ynormal;

float
adc0,
adc1,
adc0_filtered,
adc1_filtered,
x,
y;


// ----- Shaft Angle Calculations
float Angle;                                            // Raw angle ... atan2() produces 360 degrees per half-turn

int quadrantNumber = 0;                                 // quadrant IDs
int previousQuadrantNumber = 0;                         // these are used for tracking the numberOfTurns

float numberOfTurns = 0;                                // number of turns
float startAngle = 0;                                   // starting angle
float taredAngle = 0;                                   // tared angle - based on the startup value
float totalAngle = 0;                                   // total absolute angular displacement
float previousTotalAngle = 0;                           // for the display printing

// ----- Display
unsigned long Timer;
const unsigned long Next_sample = 15L;

// ----- Low Pass Filter
// See https://www.youtube.com/watch?v=eM4VHtettGg&list=PL3BbZZ9thFk5eGKN0-Xo64vwgZQk_sJmB&index=4

template <int order> // order is 1 or 2
class LowPass
{
  private:
    float a[order];
    float b[order + 1];
    float omega0;
    float dt;
    bool adapt;
    float tn1 = 0;
    float x[order + 1]; // Raw values
    float y[order + 1]; // Filtered values

  public:
    LowPass(float f0, float fs, bool adaptive) {
      // f0: cutoff frequency (Hz)
      // fs: sample frequency (Hz)
      // adaptive: boolean flag, if set to 1, the code will automatically set
      // the sample frequency based on the time history.

      omega0 = 6.28318530718 * f0;
      dt = 1.0 / fs;
      adapt = adaptive;
      tn1 = -dt;
      for (int k = 0; k < order + 1; k++) {
        x[k] = 0;
        y[k] = 0;
      }
      setCoef();
    }

    void setCoef() {
      if (adapt) {
        float t = micros() / 1.0e6;
        dt = t - tn1;
        tn1 = t;
      }

      float alpha = omega0 * dt;
      if (order == 1) {
        a[0] = -(alpha - 2.0) / (alpha + 2.0);
        b[0] = alpha / (alpha + 2.0);
        b[1] = alpha / (alpha + 2.0);
      }
      if (order == 2) {
        float c1 = 2 * sqrt(2) / alpha;
        float c2 = 4 / (alpha * alpha);
        float denom = 1.0 + c1 + c2;
        b[0] = 1.0 / denom;
        b[1] = 2.0 / denom;
        b[2] = b[0];
        a[0] = -(2.0 - 2.0 * c2) / denom;
        a[1] = -(1.0 - c1 + c2) / (1.0 + c1 + c2);
      }
    }

    float filt(float xn) {
      // Provide me with the current raw value: x
      // I will give you the current filtered value: y
      if (adapt) {
        setCoef(); // Update coefficients if necessary
      }
      y[0] = 0;
      x[0] = xn;
      // Compute the filtered values
      for (int k = 0; k < order; k++) {
        y[0] += a[k] * y[k + 1] + b[k] * x[k];
      }
      y[0] += b[order] * x[order];

      // Save the historical values
      for (int k = order; k > 0; k--) {
        y[k] = y[k - 1];
        x[k] = x[k - 1];
      }

      // Return the filtered value
      return y[0];
    }
};

// Filter instance
LowPass<2> lp0(3, 1e3, true);
LowPass<2> lp1(3, 1e3, true);
// ----------------
// setup()
// ----------------
void setup()
{
  Serial.begin(115200);                                 // start serial

  // ----- flush the buffers (needed by some processors)
  Serial.flush();                                       // clear TX buffer
  while (Serial.available()) Serial.read();             // clear RX buffer

  // ----- initialise the ADCs
  pinMode(ADC0, INPUT);
  pinMode(ADC1, INPUT);

  // ----- initialise motor
  pinMode(DIR1, OUTPUT);                                // initialise motor1
  pinMode(STEP1, OUTPUT);
  digitalWrite(DIR1, CW);
  delayMicroseconds(PULSE_WIDTH);
  digitalWrite(STEP1, LOW);

  // ----- Initialise max/min values
  Xmax = -10000.0;
  Xmin = 10000.0;
  Ymax = -10000.0;
  Ymin = 10000.0;

  // ----- Initialise digital filter (
  for (int count = 0; count < 3200; count++) {
    step_motor(CW);
    adc0 = (float)analogRead(ADC0);
    adc1 = (float)analogRead(ADC1);
    adc0_filtered = lp0.filt(adc0);
    adc1_filtered = lp1.filt(adc1);
  }

  // ----- Find max/min values
  for (int count = 0; count < 3200; count++) {
    step_motor(CW);
    adc0 = (float)analogRead(ADC0);
    adc1 = (float)analogRead(ADC1);
    adc0_filtered = lp0.filt(adc0);
    adc1_filtered = lp1.filt(adc1);

    // ----- Obtain Xpeak & Xmid values
    Xmax = max(Xmax, adc0_filtered);
    Xmin = min(Xmin, adc0_filtered);
    Ymax = max(Ymax, adc1_filtered);
    Ymin = min(Ymin, adc1_filtered);
  }

  // ----- Calculate scaling factors
  Xmid = (Xmax + Xmin) / 2;
  Xpeak = Xmax - Xmid;
  Ymid = (Ymax + Ymin) / 2;
  Ypeak = Ymax - Ymid;
  Ynormal = Xpeak / Ypeak;

  // ----- Record start angle
  Angle = raw_angle();                        // Get raw angle
  tareAngle();                                // Zero it.

  // ----- Start Display Timer
  Timer = millis() + Next_sample;
}

// ----------------
// loop()
// ----------------
void loop()
{
  step_motor(CW);
  Angle = raw_angle();
  tareAngle();
  checkQuadrant();

  if (millis() >= Timer) {
    Timer += Next_sample;
    Serial.print(x);
    Serial.print("\t");
    Serial.print(y);
    Serial.print("\t");
    Serial.print(Angle / 10);            // Keep sawtooth inside range of sinewaves
    Serial.print("\t");
    Serial.println(totalAngle / 20);       // Keep sawtooth inside range of sinewaves
  }
}

// ------------------
// raw_angle()                          // Measures the raw angle ... two sine waves over 360 degrees
// ------------------
float raw_angle() {
  adc0 = (float)analogRead(ADC0);
  adc1 = (float)analogRead(ADC1);

  // ----- Compute the filtered signal
  adc0_filtered = lp0.filt(adc0);
  adc1_filtered = lp1.filt(adc1);

  // ----- Calculate the true X & Y values
  x = adc0_filtered - Xmid;
  y = (adc1_filtered - Ymid) * Ynormal;

  // ----- Calculate the absolute angle
  float angle = atan2(y, x) * RAD_TO_DEG;
  if (angle < 0) angle += 360;            // Needed as atan2 outputs negative values between 180~360 degrees

  return angle;
}

// ----------------
// tareAngle()
// ----------------
void tareAngle()
{
  // ----- recalculate angle
  taredAngle = Angle - startAngle;                         //this tares the position

  if (taredAngle < 0)                                         //if the calculated angle is negative, we need to "normalize" it
  {
    taredAngle = taredAngle + 360;                            //correction for negative numbers (i.e. -15 becomes +345)
  }
}

// ----------------
// checkQuadrant()
// ----------------
void checkQuadrant()
{
  /*
    //Quadrants:
    4  |  1
    ---|---
    3  |  2
  */

  // ----- Quadrant 1
  if (taredAngle >= 0 && taredAngle <= 90) {
    quadrantNumber = 1;
  }

  //Quadrant 2
  if (taredAngle > 90 && taredAngle <= 180) {
    quadrantNumber = 2;
  }

  //Quadrant 3
  if (taredAngle > 180 && taredAngle <= 270) {
    quadrantNumber = 3;
  }

  // ----- Quadrant 4
  if (taredAngle > 270 && taredAngle < 360) {
    quadrantNumber = 4;
  }

  if (quadrantNumber != previousQuadrantNumber)                   // if we changed quadrant
  {
    if (quadrantNumber == 1 && previousQuadrantNumber == 4) {
      numberOfTurns++;                                            // 4 --> 1 transition: CW rotation
    }

    if (quadrantNumber == 4 && previousQuadrantNumber == 1) {
      numberOfTurns--;                                            // 1 --> 4 transition: CCW rotation
    }

    previousQuadrantNumber = quadrantNumber;                      //update to the current quadrant
  }

  if (numberOfTurns > 1) numberOfTurns = 0;
  totalAngle = (numberOfTurns * 360) + taredAngle;            //number of turns (+/-) plus the actual angle within the 0-360 range
}

// ---------------------
// STEP MOTOR
// ---------------------
void step_motor(bool dir) {

  // ----- locals
  byte patternB = PORTB;
  enum {step1, dir1};                                         // Define bit positions for PORTB

  CLR(patternB, step1);                                       // Disable motor
  (dir == CCW) ? CLR(patternB, dir1) : SET(patternB, dir1);   // Set motor direction
  PORTB = patternB;
  delayMicroseconds(PULSE_WIDTH);                             // Wait for direction line to stabilise

  SET(patternB, step1);                                       // Prepare step pulse
  PORTB = patternB;                                           // Step the motor
  delayMicroseconds(PULSE_WIDTH);
  CLR(patternB, step1);                                       // Generate trailing-edge
  PORTB = patternB;                                           // Step the motors
  delayMicroseconds(DELAY_MIN);
}
